#include "pw/scene/entity.hpp"
#include "pw/scene/scene.hpp"

#include "pw/core/debug.hpp"

namespace pw {

namespace detail {

// entity relationsships are hidden
struct relationship {
    relationship() = default;
    relationship(const relationship&) = default;

    std::vector<entt::entity> children;
    entt::entity parent { entt::null };
};

}

entity::entity(scene &s)
    : _registry(s._registry)
    , _entity(s._registry ? _registry->create() : entt::null)
{
}

bool entity::add_child(entity &child)
{
    // not mixing scenes (yet)
    if (child._registry != this->_registry) {
        debug::e() << "Cannot mix entities from different scenes";
        return false;
    }

    // TODO: check circular dependencies

    // declare child relationship
    auto& r = get_or_create_component<detail::relationship>();
    r.children.push_back(child._entity);

    // declare parent
    auto& p_r = child.get_or_create_component<detail::relationship>();
    p_r.parent = _entity;

    return true;
}

bool entity::remove_child(entity& child)
{
    // not mixing scenes (yet)
    if (child._registry != this->_registry) {
        debug::e() << "Cannot mix entities from different scenes";
        return false;
    }

    // both need to have a relationship component
    if (has_component<detail::relationship>() &&
            child.has_component<detail::relationship>())
    {
        // we need to check if the child is related to the parent
        auto r_p = child.get_component<detail::relationship>();
        if (r_p.parent == _entity)
        {
            // now go ahead delete parent
            r_p.parent = entt::null;

            // now remove all children
            auto& r = this->get_component<detail::relationship>();
            r.children.erase(std::remove(r.children.begin(),r.children.end(),child._entity));

            // flag success
            return true;
        }
    }
    return false;
}

size_t entity::child_count() const
{
    return has_component<detail::relationship>() ?
                get_component<detail::relationship>().children.size() :
                0;
}

entity::~entity()
{
    this->destroy();
}

void entity::destroy()
{
    debug::d() << __PRETTY_FUNCTION__;
    if (valid()) _registry->destroy(_entity);
}

}

